import typing
import functools
import logging

logger = logging.getLogger('aioproperty')


def await_if_needed(foo):

    @functools.wraps(foo)
    async def wrapper(*args, **kwargs):
        ret = foo(*args, **kwargs)
        if isinstance(ret, typing.Awaitable):
            ret = await ret
        return ret

    return wrapper



import typing
import asyncio
from copy import copy
import warnings
import inspect
from functools import partial
from contextlib import AbstractAsyncContextManager, AbstractContextManager, AsyncExitStack
from dataclasses import dataclass, field
from pro_lambda import pro_lambda
from pro_lambda.tools import ClsInitMeta, cls_init
from pro_lambda import consts as pl_consts
import logging
from . import tools

logger = logging.getLogger('aioproperty')
_trigger_type = typing.Union[asyncio.Condition, typing.Callable[[], typing.Awaitable]]


class _ContextCounter(AbstractContextManager):

    def __init__(self):
        self._count = 0
        self._trigger = asyncio.Event()
        self._trigger.set()

    def acquire(self):
        self._count += 1
        if self._trigger.is_set():
            self._trigger.clear()

    def release(self):
        self._count -= 1
        if self._count == 0:
            self._trigger.set()

    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()

    def __await__(self):
        return (self._trigger.wait()).__await__()


class _PropContext(AbstractAsyncContextManager):
    """
    Special context-manager. Everytime someone enters
    """
    def __init__(self):
        self._contexts: typing.List[_ContextCounter] = list()

    async def __aenter__(self):
        _context = _ContextCounter()
        self._contexts.append(_context)
        return _context

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        context = self._contexts.pop()
        await context

    def acquire(self) -> _ContextCounter:
        """
        Get current _ContextCounter and acquires it

        Returns: _ContextCounter

        """
        try:
            context = self._contexts[-1]
        except IndexError:
            context = _ContextCounter()
        context.acquire()
        return context


async_context = _PropContext()


@dataclass
class _PropertyMeta(metaclass=ClsInitMeta):
    """
    Captures task from aioproperty. When awaited, return it's result.
    It also captures getter from aioproperty and one can call it using next()

    Note:
        when use next(), it returns awaitable, so if you need some math, you can do like that:

        >>> await next(prop + 1)

    Args:
        task: pro_lambda, must return Task
        getter: pro_lambda, must return new _PropertyMeta
    """
    prop: 'aioproperty'
    instance: object
    foo: pro_lambda = field(default=pro_lambda(lambda x: x), init=False)
    _others: typing.List['aioproperty'] = field(default_factory=list, init=False)

    @cls_init
    def _add_maths(cls):
        def set_foo(params):
            name, op = params

            def wrapper(self: '_PropertyMeta', other=None):
                ret = copy(self)
                ret._others = copy(ret._others)
                if isinstance(other, _PropertyMeta):
                    if other not in ret._others:
                        ret._others.append(other)

                ret.foo = op(self.foo, other)

                return ret

            setattr(cls, name, wrapper)

        list(map(set_foo, pl_consts.ops))

    def __await__(self):
        try:
            async def wrap():
                ret = getattr(self.instance, f'_{self.prop._name}', self.prop._default)
                if isinstance(ret, typing.Awaitable):
                    ret = await ret
                if self.foo.is_async:
                    return await self.foo(ret)
                else:
                    return self.foo(ret)
            return wrap().__await__()
        except Exception:
            logger.exception(f'awaiting {self.instance}.{self.prop._name}')
            raise RuntimeError()

    def add_callback(self, foo):
        foo = tools.await_if_needed(foo)
        _wrapper = tools.await_if_needed(self.foo)

        async def wrap(prop, value):
            if prop is self.prop:
                await foo(await _wrapper(value))
            else:
                await foo(await self)

        for x in [self, *self._others]:
            x.prop.add_callback(x.instance, wrap)
        return foo


def rule(prop: _PropertyMeta):
    def deco(foo):
        prop.add_callback(foo)
        return foo
    return deco


class aioproperty:

    """
    Асинхронный property.

    У async_property нет getter в привычном смысле, у него есть асинхронный setter, который возвращает
    значение, а геттер - это по сути ожидающая выполнения сеттера задача. Соответвенно, значения мы присваиваем в обычном
    синхронном режиме (за кадром выполняется create_task), а получаем значения в асинхронном.

    Это очень крутая особенность, которая станет основной киллер-фичей нашего фреймворка. Все статусы наших объектов
    асинхронны - это значит, чтобы получить его значение, нужно использовать await, вот пример:
    >>> async def run():
    >>> class SomeClass:
    ...     @aioproperty(default='hello')
    ...     async def some_property(self, value):
    ...         await asyncio.sleep(1)
    ...         return value
    >>> test = SomeClass()
    >>> print(await test.some_property)
    hello
    >>> test.some_property = 'byby'
    >>> test.some_property = 'hello'
    >>> print(await test.some_property)
    byby

    Args:
        default: дефолтное значение, возвращается, когда ни разу не вызывался setter
        default_factory: если нет дефолтного значения и сеттер еще не вызывался, то возвращается ф
    """

    @property
    def reducers(self):
        """
        Returns iterator of reducers used by this property
        """
        for _, x in self._reducers:
            yield x

    def __init__(
            self,
            setter = None,
            *,
            default=None,
            default_factory=None,
            format='{0}',
            reducers=None,
            name=None,
    ):

        self._default = default if default is not None else default_factory() if default_factory is not None else None
        self._owner = None
        self._name = name or None
        self._format = format
        self._reducers = reducers or []
        self._context_lck = asyncio.Lock()
        self._root = self
        if setter is not None:
            self(setter)

        @self.chain(priority=-1000)
        async def _process_callbacks(instance, value):
            callbacks = self.get_callbacks(instance)

            async def trigger():
                await asyncio.gather(*[x(self, value) for x in callbacks])

            asyncio.create_task(trigger())

    def __set_name__(self, owner, name):
        self._name = self._name or name
        self._owner = owner

    def __eq__(self, other):
        if isinstance(other, aioproperty):
            return other._root is self._root

    def __copy__(self):
        newone = type(self)()
        newone.__dict__.update(self.__dict__)
        newone._reducers = copy(newone._reducers)
        return newone

    def __call__(self, setter):
        self._add_reducer(setter)
        return self

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            try:
                getattr(instance, f'_{self._name}')
            except AttributeError:
                ret = asyncio.Future()
                ret.set_result(self._default)
                setattr(instance, f'_{self._name}', ret)
            ret = _PropertyMeta(
                prop=self,
                instance=instance,
            )
            return ret

    def _add_reducer(self, reducer, priority=None, last=True):
        if reducer in self.reducers:
            warnings.warn(f'attempt to add reducer twice to {self._owner}.{self._name}')
            return
        sig = inspect.signature(reducer)
        if len(sig.parameters) not in [1,2]:
            raise RuntimeError(f'reducer {self}.{reducer} has wrong parameters: {sig.parameters}, must '
                               f'have "self, value" or just "value"')

        @tools.await_if_needed
        def wrap_reducer(instance=None, value=None):
            if len(sig.parameters) == 1:
                return reducer(value)
            elif len(sig.parameters) == 2:
                return reducer(instance, value)

        if priority is None:
            if self._reducers:
                if last:
                    priority = max([x[0] for x in self._reducers]) + 1
                else:
                    priority = min([x[0] for x in self._reducers]) - 1
            else:
                priority = 0

        wrap_reducer.__name__ = reducer.__name__
        self._reducers.append((priority, wrap_reducer))
        self._reducers.sort(key=lambda x: x[0])

    async def _reduce(self, instance, value):
        """
        Основной редьюсер
        Args:
            instance:
            value:
        """
        for reducer in self.reducers:
            value = await reducer(instance, value) or value
        return value

    def get_callbacks(self, instance) -> typing.List[typing.Callable]:
        try:
            _trig = getattr(instance, f'_t_{self._name}')
        except AttributeError:
            _trig = []
            setattr(instance, f'_t_{self._name}', _trig)
        return _trig

    def add_callback(self, instance, callback: typing.Callable):
        callbacks = self.get_callbacks(instance)
        callbacks.append(callback)

    def __set__(self, instance, value):
        try:
            prev_task: asyncio.Future = getattr(instance, f'_{self._name}')
        except AttributeError:
            prev_task = asyncio.Future()
            prev_task.set_result(self._default)

        _context = async_context.acquire()

        async def wrap():
            prev_value = await prev_task
            try:
                if value != prev_value:
                    logger.debug(f'set {instance}.{self._name}={value}')
                    try:
                        return await self._reduce(instance, value)
                    except Exception as exc:
                        logger.exception(f'error during setting {instance}.{self._name} to {value}')
                        return prev_value
                else:
                    return value
            finally:
                _context.release()

        setattr(instance, f'_{self._name}', asyncio.create_task(wrap()))

    async def _default_setter(self, instance, value):
        return value

    def chain(self, _foo = None, *, is_first=True, priority=None):
        """
        Добавляет функцию в цепочку редьюсеров
        Args:
            is_first: если истина, добавляется в начало, в противном случае в конец
            priority: можно управлять последовательностью исполнения редьюсеров. Чем меньше значение, тем раньше он будет
                запущен
        Returns:

        """
        def deco(foo):
            self._add_reducer(foo, last=not is_first, priority=priority)
            return foo

        return deco(_foo) if _foo is not None else deco


class inject:

    def __init__(
            self,
            parent_prop: typing.Union[aioproperty, str],
            *,
            priority=None,
            is_first=True,
    ):
        """
        Decorator, decorated foo will be chained to parent's property, but in new class

        Args:
            parent_prop: can be aioproperty or string-name of property to chain new foo
            priority: like in aioproperty.chain
            is_first: like in aioproperty.chain

        Notes:
            after decoration, prop becomes actually a copy of a previous prop. So you can not
            `prop1 is prop2 == False`, if you need subclass-comparing you can use `prop1 == prop2`
        """
        if isinstance(parent_prop, aioproperty):
            self.name = parent_prop._name
        else:
            self.name = parent_prop
        self.foo = None
        self.priority=priority
        self.is_first = is_first

    def __call__(self, foo):
        self.foo = foo
        return self

    def __set_name__(self, owner, name):
        _new = copy(getattr(owner, self.name))
        _new._owner = owner
        _new.chain(self.foo, priority=self.priority, is_first=self.is_first)
        setattr(owner, self.name, _new)



import pytest
from aioproperty import aioproperty, async_context, rule, inject
from pytest import fixture
import asyncio


class SomeClass:

    res_test_prop = 0
    res_test_prop_2 = 0
    res_test_prop_3 = 0

    def __init__(self):

        self.res_col_1 = []
        self.res_col_2 = []
        self.res_col_3 = []

    @aioproperty(default=0)
    async def test_prop(self, value):
        await asyncio.sleep(0.1)
        self.res_col_1.append(value)

    @test_prop.chain
    async def some_hook(self, value):
        print('hook1', value)
        self.res_test_prop = value

    @aioproperty(default=0)
    async def test_prop_2(self, value):
        await asyncio.sleep(1)
        self.res_test_prop_2 = value
        self.res_col_2.append(value)

    @aioproperty(default=0)
    async def test_prop_3(self, value):
        self.res_test_prop_3 = value
        self.res_col_3.append(value)


class SomeInherited(SomeClass):

    def __init__(self):
        super().__init__()
        self.recieve_injection = []
        self.recieve_another_injection = []

    @inject(SomeClass.test_prop)
    async def add_some_more(self, value):
        self.recieve_injection.append(value)

    @inject('test_prop')
    async def add_some_more_2(self, value):
        self.recieve_another_injection.append(value + 1)


@fixture(scope='session')
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()


@fixture()
async def some_obj():
    yield SomeClass()

@fixture()
async def waiting_task(some_obj):
    collection = []

    @rule(some_obj.test_prop)
    def some_rule(value):
        collection.append(value)

    yield collection


@fixture()
async def waiting_task_2(some_obj):
    collection = []

    @rule(some_obj.test_prop + some_obj.test_prop_2)
    def some_rule(value):
        collection.append(value)

    yield collection


@pytest.mark.asyncio
async def test_1(some_obj, waiting_task, waiting_task_2):
    assert await some_obj.test_prop == 0
    async with async_context:
        async with async_context:
            some_obj.test_prop_3 = 1
        some_obj.test_prop = 1
        some_obj.test_prop_2 = 1

    assert some_obj.res_test_prop == 1
    assert some_obj.res_test_prop_2 == 1
    assert some_obj.res_test_prop_3 == 1

    assert await some_obj.test_prop == 1
    assert await some_obj.test_prop_2 == 1
    assert await some_obj.test_prop_3 == 1

    assert await (some_obj.test_prop + 1) == 2
    assert await ((some_obj.test_prop + some_obj.test_prop_2) / 2) == 1

    some_obj.test_prop_2 = 3
    assert await ((some_obj.test_prop + some_obj.test_prop_2) / 2) == 2
    assert waiting_task == [1]
    async with async_context:
        some_obj.test_prop = 2


@pytest.mark.asyncio
async def test_multiple(some_obj):
    async with async_context:
        some_obj.test_prop_3 = 1
        some_obj.test_prop_3 = 2
        some_obj.test_prop_2 = 3
        some_obj.test_prop = 4
        some_obj.test_prop_3 = 5
    assert some_obj.res_col_3 == [1,2,5]
    assert some_obj.res_col_1 == [4]
    assert some_obj.res_col_2 == [3]


@pytest.mark.asyncio
async def test_triggers(waiting_task, waiting_task_2, some_obj):
    async with async_context:
        some_obj.test_prop = 1
        some_obj.test_prop = 3
    async with async_context:
        some_obj.test_prop_2 = 3
    async with async_context:
        some_obj.test_prop = 2
    assert waiting_task == [1, 3, 2]
    assert waiting_task_2 == [1, 3, 6, 5]


@fixture()
async def some_other_obj():
    yield SomeInherited()


@pytest.mark.asyncio
async def test_inherited(some_other_obj, some_obj):
    async with async_context:
        some_obj.test_prop = 1
        some_other_obj.test_prop = 1

    assert some_obj.res_col_1 == [1]
    assert some_other_obj.res_col_1 == [1]
    assert some_other_obj.recieve_injection == [1]
    assert some_other_obj.recieve_another_injection == [2]
    assert SomeInherited.test_prop == SomeClass.test_prop


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = 'aioproperty'
copyright = '2019, Viktorov A.G.'
author = 'Viktorov A.G.'

# The full version, including alpha/beta/rc tags
release = '0.2.1'
master_doc = 'index'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx_rtd_theme',
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
    'sphinx.ext.napoleon',
    'm2r'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


html_theme_options = {
    'canonical_url': '',
    'analytics_id': 'UA-XXXXXXX-1',  #  Provided by Google in your dashboard
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    'style_nav_header_background': 'white',
    # Toc options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': False
}



from setuptools import setup

with open('README.md', 'r') as f:
    desc = f.read()

setup(
    name='aioproperty',
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    python_requires='>=3.7',
    packages=['aioproperty'],
    url='',
    license='MIT',
    author='Viktorov A.G.',
    author_email='andvikt@gmail.com',
    description='truely async properties',
    long_description=desc,
    long_description_content_type='text/markdown',
    install_requires=['pro_lambda>=0.3.3'],
    tests_require=['pytest', 'pytest-asyncio']
)
